"""Terminal capability builder patterns."""
# std imports
import re
from collections import OrderedDict

__all__ = (
    'CAPABILITY_DATABASE',
    'CAPABILITIES_RAW_MIXIN',
    'CAPABILITIES_ADDITIVES',
    'CAPABILITIES_CAUSE_MOVEMENT',
)

CAPABILITY_DATABASE = OrderedDict((
    ('bell', ('bel', {})),
    ('carriage_return', ('cr', {})),
    ('change_scroll_region', ('csr', {'nparams': 2})),
    ('clear_all_tabs', ('tbc', {})),
    ('clear_screen', ('clear', {})),
    ('clr_bol', ('el1', {})),
    ('clr_eol', ('el', {})),
    ('clr_eos', ('clear_eos', {})),
    ('column_address', ('hpa', {'nparams': 1})),
    ('cursor_address', ('cup', {'nparams': 2, 'match_grouped': True})),
    ('cursor_down', ('cud1', {})),
    ('cursor_home', ('home', {})),
    ('cursor_invisible', ('civis', {})),
    ('cursor_left', ('cub1', {})),
    ('cursor_normal', ('cnorm', {})),
    ('cursor_report', ('u6', {'nparams': 2, 'match_grouped': True})),
    ('cursor_right', ('cuf1', {})),
    ('cursor_up', ('cuu1', {})),
    ('cursor_visible', ('cvvis', {})),
    ('delete_character', ('dch1', {})),
    ('delete_line', ('dl1', {})),
    ('enter_blink_mode', ('blink', {})),
    ('enter_bold_mode', ('bold', {})),
    ('enter_dim_mode', ('dim', {})),
    ('enter_fullscreen', ('smcup', {})),
    ('enter_standout_mode', ('standout', {})),
    ('enter_superscript_mode', ('superscript', {})),
    ('enter_susimpleript_mode', ('susimpleript', {})),
    ('enter_underline_mode', ('underline', {})),
    ('erase_chars', ('ech', {'nparams': 1})),
    ('exit_alt_charset_mode', ('rmacs', {})),
    ('exit_am_mode', ('rmam', {})),
    ('exit_attribute_mode', ('sgr0', {})),
    ('exit_ca_mode', ('rmcup', {})),
    ('exit_fullscreen', ('rmcup', {})),
    ('exit_insert_mode', ('rmir', {})),
    ('exit_standout_mode', ('rmso', {})),
    ('exit_underline_mode', ('rmul', {})),
    ('flash_hook', ('hook', {})),
    ('flash_screen', ('flash', {})),
    ('insert_line', ('il1', {})),
    ('keypad_local', ('rmkx', {})),
    ('keypad_xmit', ('smkx', {})),
    ('meta_off', ('rmm', {})),
    ('meta_on', ('smm', {})),
    ('orig_pair', ('op', {})),
    ('parm_down_cursor', ('cud', {'nparams': 1})),
    ('parm_left_cursor', ('cub', {'nparams': 1, 'match_grouped': True})),
    ('parm_dch', ('dch', {'nparams': 1})),
    ('parm_delete_line', ('dl', {'nparams': 1})),
    ('parm_ich', ('ich', {'nparams': 1})),
    ('parm_index', ('indn', {'nparams': 1})),
    ('parm_insert_line', ('il', {'nparams': 1})),
    ('parm_right_cursor', ('cuf', {'nparams': 1, 'match_grouped': True})),
    ('parm_rindex', ('rin', {'nparams': 1})),
    ('parm_up_cursor', ('cuu', {'nparams': 1})),
    ('print_screen', ('mc0', {})),
    ('prtr_off', ('mc4', {})),
    ('prtr_on', ('mc5', {})),
    ('reset_1string', ('r1', {})),
    ('reset_2string', ('r2', {})),
    ('reset_3string', ('r3', {})),
    ('restore_cursor', ('rc', {})),
    ('row_address', ('vpa', {'nparams': 1})),
    ('save_cursor', ('sc', {})),
    ('scroll_forward', ('ind', {})),
    ('scroll_reverse', ('rev', {})),
    ('set0_des_seq', ('s0ds', {})),
    ('set1_des_seq', ('s1ds', {})),
    ('set2_des_seq', ('s2ds', {})),
    ('set3_des_seq', ('s3ds', {})),
    # this 'color' is deceiving, but often matching, and a better match
    # than set_a_attributes1 or set_a_foreground.
    ('color', ('_foreground_color', {'nparams': 1, 'match_any': True,
                                     'numeric': 1})),
    ('set_a_foreground', ('color', {'nparams': 1, 'match_any': True,
                                    'numeric': 1})),
    ('set_a_background', ('on_color', {'nparams': 1, 'match_any': True,
                                       'numeric': 1})),
    ('set_tab', ('hts', {})),
    ('tab', ('ht', {})),
    ('italic', ('sitm', {})),
    ('no_italic', ('sitm', {})),
))

CAPABILITIES_RAW_MIXIN = {
    'bell': re.escape('\a'),
    'carriage_return': re.escape('\r'),
    'cursor_left': re.escape('\b'),
    'cursor_report': re.escape('\x1b') + r'\[(\d+)\;(\d+)R',
    'cursor_right': re.escape('\x1b') + r'\[C',
    'exit_attribute_mode': re.escape('\x1b') + r'\[m',
    'parm_left_cursor': re.escape('\x1b') + r'\[(\d+)D',
    'parm_right_cursor': re.escape('\x1b') + r'\[(\d+)C',
    'restore_cursor': re.escape(r'\x1b\[u'),
    'save_cursor': re.escape(r'\x1b\[s'),
    'scroll_forward': re.escape('\n'),
    'set0_des_seq': re.escape('\x1b(B'),
    'tab': re.escape('\t'),
}
_ANY_NOTESC = '[^' + re.escape('\x1b') + ']*'

CAPABILITIES_ADDITIVES = {
    'link': ('link',
             re.escape('\x1b') + r'\]8;' + _ANY_NOTESC + ';' +
             _ANY_NOTESC + re.escape('\x1b') + '\\\\'),
    'color256': ('color', re.escape('\x1b') + r'\[38;5;\d+m'),
    'on_color256': ('on_color', re.escape('\x1b') + r'\[48;5;\d+m'),
    'color_rgb': ('color_rgb', re.escape('\x1b') + r'\[38;2;\d+;\d+;\d+m'),
    'on_color_rgb': ('on_color_rgb', re.escape('\x1b') + r'\[48;2;\d+;\d+;\d+m'),
    'shift_in': ('', re.escape('\x0f')),
    'shift_out': ('', re.escape('\x0e')),
    # sgr(...) outputs strangely, use the basic ANSI/EMCA-48 codes here.
    'set_a_attributes1': (
        'sgr', re.escape('\x1b') + r'\[\d+m'),
    'set_a_attributes2': (
        'sgr', re.escape('\x1b') + r'\[\d+\;\d+m'),
    'set_a_attributes3': (
        'sgr', re.escape('\x1b') + r'\[\d+\;\d+\;\d+m'),
    'set_a_attributes4': (
        'sgr', re.escape('\x1b') + r'\[\d+\;\d+\;\d+\;\d+m'),
    # this helps where xterm's sgr0 includes set0_des_seq, we'd
    # rather like to also match this immediate substring.
    'sgr0': ('sgr0', re.escape('\x1b') + r'\[m'),
    'backspace': ('', re.escape('\b')),
    'ascii_tab': ('', re.escape('\t')),
    'clr_eol': ('', re.escape('\x1b[K')),
    'clr_eol0': ('', re.escape('\x1b[0K')),
    'clr_bol': ('', re.escape('\x1b[1K')),
    'clr_eosK': ('', re.escape('\x1b[2K')),
}

CAPABILITIES_CAUSE_MOVEMENT = (
    'ascii_tab',
    'backspace',
    'carriage_return',
    'clear_screen',
    'column_address',
    'cursor_address',
    'cursor_down',
    'cursor_home',
    'cursor_left',
    'cursor_right',
    'cursor_up',
    'enter_fullscreen',
    'exit_fullscreen',
    'parm_down_cursor',
    'parm_left_cursor',
    'parm_right_cursor',
    'parm_up_cursor',
    'restore_cursor',
    'row_address',
    'scroll_forward',
    'tab',
)



# std imports
from typing import Any, Dict, Tuple, OrderedDict

CAPABILITY_DATABASE: OrderedDict[str, Tuple[str, Dict[str, Any]]]
CAPABILITIES_RAW_MIXIN: Dict[str, str]
CAPABILITIES_ADDITIVES: Dict[str, Tuple[str, str]]
CAPABILITIES_CAUSE_MOVEMENT: Tuple[str, ...]



"""Configure test fixtures"""

# std imports
import os
import platform
import subprocess

# 3rd party
import pytest

IS_WINDOWS = platform.system() == 'Windows'

all_terms_params = 'xterm screen ansi vt220 rxvt cons25 linux'.split()
many_lines_params = [40, 80]
# we must test a '1' column for conditional in _handle_long_word
many_columns_params = [1, 10]


def envvar_enabled(envvar):
    """
    Return True if environment variable is set and enabled

    unset values, 'no', 0, and 'false' and treated as False regardless of case
    All other values are considered True
    """

    value = os.environ.get(envvar, False)

    if value is False:
        return value

    if value.lower() in ('no', 'false'):
        return False

    try:
        return bool(int(value))
    except ValueError:
        return True


TEST_FULL = envvar_enabled('TEST_FULL')
TEST_KEYBOARD = envvar_enabled('TEST_KEYBOARD')
TEST_QUICK = envvar_enabled('TEST_QUICK')
TEST_RAW = envvar_enabled('TEST_RAW')


if TEST_FULL:
    try:
        all_terms_params = [
            # use all values of the first column of data in output of 'toe -a'
            _term.split(None, 1)[0] for _term in
            subprocess.Popen(('toe', '-a'),  # pylint: disable=consider-using-with
                             stdout=subprocess.PIPE,
                             close_fds=True)
            .communicate()[0].splitlines()]
    except OSError:
        pass
elif IS_WINDOWS:
    all_terms_params = ['vtwin10', ]
elif TEST_QUICK:
    all_terms_params = 'xterm screen ansi linux'.split()


if TEST_QUICK:
    many_lines_params = [80, ]
    many_columns_params = [25, ]


@pytest.fixture(params=all_terms_params)
def all_terms(request):
    """Common kind values for all kinds of terminals."""
    return request.param


@pytest.fixture(params=many_lines_params)
def many_lines(request):
    """Various number of lines for screen height."""
    return request.param


@pytest.fixture(params=many_columns_params)
def many_columns(request):
    """Various number of columns for screen width."""
    return request.param



# std imports
from typing import Set, Dict, Type, Mapping, TypeVar, Iterable, Optional, OrderedDict

# local
from .terminal import Terminal

_T = TypeVar("_T")

class Keystroke(str):
    def __new__(
        cls: Type[_T],
        ucs: str = ...,
        code: Optional[int] = ...,
        name: Optional[str] = ...,
    ) -> _T: ...
    @property
    def is_sequence(self) -> bool: ...
    @property
    def name(self) -> Optional[str]: ...
    @property
    def code(self) -> Optional[int]: ...

def get_keyboard_codes() -> Dict[int, str]: ...
def get_keyboard_sequences(term: Terminal) -> OrderedDict[str, int]: ...
def get_leading_prefixes(sequences: Iterable[str]) -> Set[str]: ...
def resolve_sequence(
    text: str, mapper: Mapping[str, int], codes: Mapping[int, str]
) -> Keystroke: ...



# -*- coding: utf-8 -*-
"""Module containing Windows version of :class:`Terminal`."""

from __future__ import absolute_import

# std imports
import time
import msvcrt  # pylint: disable=import-error
import contextlib

# 3rd party
from jinxed import win32  # pylint: disable=import-error

# local
from .terminal import WINSZ
from .terminal import Terminal as _Terminal


class Terminal(_Terminal):
    """Windows subclass of :class:`Terminal`."""

    def getch(self):
        r"""
        Read, decode, and return the next byte from the keyboard stream.

        :rtype: unicode
        :returns: a single unicode character, or ``u''`` if a multi-byte
            sequence has not yet been fully received.

        For versions of Windows 10.0.10586 and later, the console is expected
        to be in ENABLE_VIRTUAL_TERMINAL_INPUT mode and the default method is
        called.

        For older versions of Windows, msvcrt.getwch() is used. If the received
        character is ``\x00`` or ``\xe0``, the next character is
        automatically retrieved.
        """
        if win32.VTMODE_SUPPORTED:
            return super(Terminal, self).getch()

        rtn = msvcrt.getwch()
        if rtn in ('\x00', '\xe0'):
            rtn += msvcrt.getwch()
        return rtn

    def kbhit(self, timeout=None):
        """
        Return whether a keypress has been detected on the keyboard.

        This method is used by :meth:`inkey` to determine if a byte may
        be read using :meth:`getch` without blocking.  This is implemented
        by wrapping msvcrt.kbhit() in a timeout.

        :arg float timeout: When ``timeout`` is 0, this call is
            non-blocking, otherwise blocking indefinitely until keypress
            is detected when None (default). When ``timeout`` is a
            positive number, returns after ``timeout`` seconds have
            elapsed (float).
        :rtype: bool
        :returns: True if a keypress is awaiting to be read on the keyboard
            attached to this terminal.
        """
        end = time.time() + (timeout or 0)
        while True:

            if msvcrt.kbhit():
                return True

            if timeout is not None and end < time.time():
                break

            time.sleep(0.01)  # Sleep to reduce CPU load
        return False

    @staticmethod
    def _winsize(fd):
        """
        Return named tuple describing size of the terminal by ``fd``.

        :arg int fd: file descriptor queries for its window size.
        :rtype: WINSZ
        :returns: named tuple describing size of the terminal

        WINSZ is a :class:`collections.namedtuple` instance, whose structure
        directly maps to the return value of the :const:`termios.TIOCGWINSZ`
        ioctl return value. The return parameters are:

            - ``ws_row``: width of terminal by its number of character cells.
            - ``ws_col``: height of terminal by its number of character cells.
            - ``ws_xpixel``: width of terminal by pixels (not accurate).
            - ``ws_ypixel``: height of terminal by pixels (not accurate).
        """
        window = win32.get_terminal_size(fd)
        return WINSZ(ws_row=window.lines, ws_col=window.columns,
                     ws_xpixel=0, ws_ypixel=0)

    @contextlib.contextmanager
    def cbreak(self):
        """
        Allow each keystroke to be read immediately after it is pressed.

        This is a context manager for ``jinxed.w32.setcbreak()``.

        .. note:: You must explicitly print any user input you would like
            displayed.  If you provide any kind of editing, you must handle
            backspace and other line-editing control functions in this mode
            as well!

        **Normally**, characters received from the keyboard cannot be read
        by Python until the *Return* key is pressed. Also known as *cooked* or
        *canonical input* mode, it allows the tty driver to provide
        line-editing before shuttling the input to your program and is the
        (implicit) default terminal mode set by most unix shells before
        executing programs.
        """
        if self._keyboard_fd is not None:

            filehandle = msvcrt.get_osfhandle(self._keyboard_fd)

            # Save current terminal mode:
            save_mode = win32.get_console_mode(filehandle)
            save_line_buffered = self._line_buffered
            win32.setcbreak(filehandle)
            try:
                self._line_buffered = False
                yield
            finally:
                win32.set_console_mode(filehandle, save_mode)
                self._line_buffered = save_line_buffered

        else:
            yield

    @contextlib.contextmanager
    def raw(self):
        """
        A context manager for ``jinxed.w32.setcbreak()``.

        Although both :meth:`break` and :meth:`raw` modes allow each keystroke
        to be read immediately after it is pressed, Raw mode disables
        processing of input and output.

        In cbreak mode, special input characters such as ``^C`` are
        interpreted by the terminal driver and excluded from the stdin stream.
        In raw mode these values are receive by the :meth:`inkey` method.
        """
        if self._keyboard_fd is not None:

            filehandle = msvcrt.get_osfhandle(self._keyboard_fd)

            # Save current terminal mode:
            save_mode = win32.get_console_mode(filehandle)
            save_line_buffered = self._line_buffered
            win32.setraw(filehandle)
            try:
                self._line_buffered = False
                yield
            finally:
                win32.set_console_mode(filehandle, save_mode)
                self._line_buffered = save_line_buffered

        else:
            yield
